use anyhow::Result;
use openssl::{
    bn::{BigNum, BigNumContext},
    ec::{EcGroup, EcPoint},
};

pub struct EllipticCurve {
    pub curve: EcGroup,
    pub gen_point: EcPoint,
    pub pub_point: EcPoint,
}

pub struct PrivateKey {
    pub gen_order: BigNum,
    pub private_key: BigNum,
}

impl PrivateKey {
    pub fn new(gen_order: &str, private_key: &str) -> Result<Self> {
        let gen_order = BigNum::from_dec_str(gen_order)?;
        let private_key = &gen_order - &BigNum::from_dec_str(private_key)?;
        Ok(Self {
            gen_order,
            private_key,
        })
    }
}

impl EllipticCurve {
    pub fn new(
        p: &str,
        a: &str,
        b: &str,
        generator_x: &str,
        generator_y: &str,
        public_key_x: &str,
        public_key_y: &str,
    ) -> Result<Self> {
        let mut context = BigNumContext::new()?;

        let p = BigNum::from_dec_str(p)?;
        let a = BigNum::from_dec_str(a)?;
        let b = BigNum::from_dec_str(b)?;
        let generator_x = BigNum::from_dec_str(generator_x)?;
        let generator_y = BigNum::from_dec_str(generator_y)?;
        let public_key_x = BigNum::from_dec_str(public_key_x)?;
        let public_key_y = BigNum::from_dec_str(public_key_y)?;

        let curve = EcGroup::from_components(p, a, b, &mut context)?;

        let mut gen_point = EcPoint::new(&curve)?;
        gen_point.set_affine_coordinates_gfp(&curve, &generator_x, &generator_y, &mut context)?;

        let mut pub_point = EcPoint::new(&curve)?;
        pub_point.set_affine_coordinates_gfp(&curve, &public_key_x, &public_key_y, &mut context)?;

        Ok(Self {
            curve,
            gen_point,
            pub_point,
        })
    }
}
